/* Demo of rotation and translation applied to planetary movement */
/* An OpenGL demo by A.P.Gardner and P.R.Miller */
/* The demo is not meant to exactly show the relative movement of the bodies involved. */
/* If you're an astronomer and can't stand the errors, you have the source code so */
/* get programming :) */

#include <GL/glut.h>
#include <stdio.h>
#include <stdlib.h>
#include "types.h"
#include "maths.h"
#include "drawing.h"

/* variables */

GLuint texName[3];			/* store of texture references */
world_vector position;		/* camera position */
orientation angle;			/* camera orientation */
int left_mouse = 0;			/* indicates left mouse button is pressed */
int right_mouse = 0;		/* indicates right mouse button is pressed */
int mouse_x;				/* last position of mouse in x axis */
int mouse_y;				/* last position of mouse in y axis */
float earth_angle = 0.0;	/* current rotation of earth */
float moon_angle = 0.0;		/* current orbital rotation of moon */

/* Function to load a texture from a bmp file */

void makeTextureImage(char filename[], int tex_name)
{
	long data_start = 54;			/* start of data section in files wall.bmp & light.bmp */
									/* NB header size of bmp files is not fixed, refer to M$ documentation */
	unsigned char ch, cho[4];		/* temp file input buffers */
	FILE *fileptr;					/* pointer to bmp file */
	GLubyte image256[256][256][3];	/* storage for 256x256 texture */
	int n, m, i, j;					/* counters */

	/* open file */

	if ((fileptr = fopen(filename, "rb")) == NULL)
		return;

	/* step over the header */

	n = 0;
	while (!feof(fileptr) && n < data_start)
	{
		ch = fgetc(fileptr);
		n++;
	}

	/* load the texture data */

	for (i=0; i<256; i++)
	{
		for (j=0; j<256; j++)
		{
			m=0;
			while(!feof(fileptr) && m < 3)
			{
				cho[m] = fgetc(fileptr);
				m++;
			}
			image256[i][j][0] = cho[2];	/* bmp using BGR values, hence the swap */
			image256[i][j][1] = cho[1];
			image256[i][j][2] = cho[0];
		}
	}

	/* close the file */

	fclose(fileptr);

	/* create texture object */
	
	glBindTexture(GL_TEXTURE_2D, texName[tex_name]);

	/* set up wrap & filter params */

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	/* load the image */

	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, 256, 256,
				 0, GL_RGB, GL_UNSIGNED_BYTE, image256);
}

/* Function to initialise params */

void init(void)
{
	GLfloat diffuse_material[] = {0.8, 0.8, 0.8, 1.0};
	GLfloat ambient_light[] = {0.1, 0.1, 0.1, 1.0};

	/* initialise data */

	position.x = -250.0; position.y = 0.0, position.z = -250.0;
	angle.pitch = 0.0; angle.yaw = 45.0; angle.roll = 0.0;

	/* generate texture names */

	glGenTextures(2, texName);

	/* load textures */

	makeTextureImage("stars.bmp", 0);
	makeTextureImage("earth.bmp", 1);
	makeTextureImage("moon.bmp", 2);

	/* set material properties */

	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, diffuse_material);

	/* set lighting params */

	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambient_light);
	glLightfv(GL_LIGHT0, GL_AMBIENT, ambient_light);
	glEnable(GL_LIGHT0);
	glEnable(GL_LIGHTING);

	/* set culling params */

	glEnable(GL_CULL_FACE);		
	glCullFace(GL_BACK);		/* cull back facing polygons */

	/* enable depth buffering */

	glEnable(GL_DEPTH_TEST);

	/* set texturing function */

	glEnable(GL_TEXTURE_2D);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
}

/* Function to clean up */

void cleanup(void)
{
	/* delete texture objects */

	glDeleteTextures(3, texName);
}

/* Function to set camera direction */

void SetCameraRotation(GLfloat yaw, GLfloat pitch)
{
	/* rotate scene */

	glRotatef(-pitch, 1.0, 0.0, 0.0);
	glRotatef(yaw, 0.0, 1.0, 0.0);
}

/* Function to set camera position */

void SetCameraPosition(GLfloat x, GLfloat y, GLfloat z)
{
	/* allow for translate scene */

	glTranslatef(-x, -y, z);
}

/* Function to draw display */

void display(void)
{
	GLfloat position_one[] = {-1.0, 0.0, 0.0, 0.0};
	
	/* update rotations */

	earth_angle += 1.0;
	if (earth_angle > 360.0)
		earth_angle -= 360.0;
	moon_angle += 1.0 / 27.3;
	if (moon_angle > 360.0)
		moon_angle -= 360.0;

	/* clear all buffers */

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	/* clear the matrix */

	glLoadIdentity();

	/* set up the viewing direction */

	SetCameraRotation(angle.yaw, angle.pitch);

	/* draw the background star field */

	glDisable(GL_LIGHTING);
	glBindTexture(GL_TEXTURE_2D, texName[0]);
	draw_skybox();

	/* draw the sun */

	glDisable(GL_TEXTURE_2D);
	glColor3f(1.0, 1.0, 0.9);
	glPushMatrix();
	glTranslatef(-500.0, 0.0, 0.0);
	draw_sphere(5.0, 5);
	glPopMatrix();
	glColor3f(1.0, 1.0, 1.0);
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_LIGHTING);

	/* set up the viewing position */

	SetCameraPosition(position.x, position.y, position.z);

	/* set direction of light, confusing eh? */
	/* because w param in position_one is zero, GL_LIGHT0 becomes a directional */
	/* light source and so this call sets direction, not position */

	glLightfv(GL_LIGHT0, GL_POSITION, position_one);

	/* draw scene */

	glPushMatrix();
		glBindTexture(GL_TEXTURE_2D,texName[1]);
		glRotatef(23.0, 0.0, 0.0, 1.0);	/* earth is tilted relative to its orbital plane */
		glRotatef(earth_angle, 0.0, 1.0, 0.0);
		draw_sphere(10.0, 10);
	glPopMatrix();
	glRotatef(moon_angle, 0.0, 1.0, 0.0);
	glTranslatef(0.0, 0.0, -185.0);		/* should be about 300.0 to be correct */
	glRotatef(10.0, 1.0, 0.0, 0.0);
	glBindTexture(GL_TEXTURE_2D,texName[2]);
	draw_sphere(2.72, 10); 

	/* swap front and back buffers */

	glutSwapBuffers();
}

/* Function to handle window size changes */

void reshape(int w, int h)
{
	glViewport(0, 0, (GLsizei)w, (GLsizei)h);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(45.0, (GLfloat)w/(GLfloat)h, 2.0, 5000.0);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

/* Function to handle key presses */

void keyboard(unsigned char key, int x, int y)
{
	world_vector dir;

	if (key == 'w' || key == 'W')
	{
		dir = orientation_to_unit_vector(angle);
		dir = world_vector_scale(dir, 0.5);
		position = world_vector_add(position, dir);
	}		

	if (key == 's' || key == 'S')
	{
		dir = orientation_to_unit_vector(angle);
		dir = world_vector_scale(dir, -0.5);
		position = world_vector_add(position, dir);
	}		

	if (key == 'a' || key == 'A')
		angle.yaw = angle.yaw - 5;

	if (key == 'd' || key == 'D')
		angle.yaw = angle.yaw + 5;

	if (key == 'z' || key == 'Z')
		angle.pitch = angle.pitch - 5;

	if (key == 'x' || key == 'X')
		angle.pitch = angle.pitch + 5;
}

/* Function to handle mouse inputs */

void mouse(int button, int state, int x, int y)
{
	/* check for mouse button state changes */

	if (button == GLUT_RIGHT_BUTTON && state == GLUT_DOWN)
		right_mouse = 1;

	if (button == GLUT_RIGHT_BUTTON && state == GLUT_UP)
		right_mouse = 0;

	if (button == GLUT_LEFT_BUTTON && state == GLUT_DOWN)
		left_mouse = 1;

	if (button == GLUT_LEFT_BUTTON && state == GLUT_UP)
		left_mouse = 0;

	/* store mouse position */

	mouse_x = x;
	mouse_y = y;
}

/* Function to handle mouse movements */

void motion(int x, int y)
{
	float mouse_move_x = 0.0, mouse_move_y = 0.0;

	/* check for movements */

	if (x != mouse_x)
	{
		mouse_move_x = x - mouse_x;
		mouse_x = x;
	}

	if (y != mouse_y)
	{
		mouse_move_y = y - mouse_y;
		mouse_y = y;
	}

	/* change orientation of camera */

	if (mouse_move_x != 0.0)
		angle.yaw = angle.yaw + mouse_move_x;

	if (mouse_move_y != 0.0)
		angle.pitch = angle.pitch + mouse_move_y;
}

/* Idle function */

void idle(void)
{
	world_vector dir;

	/* move camera if required */

	if (right_mouse)
	{
		dir = orientation_to_unit_vector(angle);
		dir = world_vector_scale(dir, 0.5);
		position = world_vector_add(position, dir);
	}		

	/* update display */

	glutPostRedisplay();
}

/* Main function */

int main(int argc, char** argv)
{
	/* print instructions */

	printf("PLANETARY MOVEMENT DEMO\n\n");
	printf("Controls:\n\n");
	printf("Left mouse button - mouse movements change view\n");
	printf("Right mouse button - move forward\n");
	printf("W - move forward\n");
	printf("S - move backward\n");
	printf("A - turn left\n");
	printf("D - turn right\n");
	printf("Z - look down\n");
	printf("X - look up\n");

	/* run the demo */

	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGB);
	glutInitWindowSize(640, 480);
	glutInitWindowPosition(100, 100);
	glutCreateWindow("OpenGL Demo - Planetary Movement");
	init();
	glutDisplayFunc(display);
	glutReshapeFunc(reshape);
	glutKeyboardFunc(keyboard);
	glutMouseFunc(mouse);
	glutMotionFunc(motion);
	glutIdleFunc(idle);
	glutMainLoop();
	cleanup();

	return 0;
}